package de.tub.mapc.knowledge;

import java.io.IOException;

import de.dailab.keks.io.IDeserialiser;
import de.dailab.keks.io.ISerialiser;
import de.jiac.micro.agent.memory.IFact;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.lang.IntHashMap;
import de.tub.mapc.worldmodel.GridField;

public class Gridworld implements IFact {

	private String id;
	private String opponent;
	private int steps;
	private int gsizex;
	private int gsizey;
	private int corralx0;
	private int corraly0;
	private int corralx1;
	private int corraly1;
	private int lineOfSight;
	private IntHashMap<Switch> switches;
	
	private GridField[][] grid;	
	
	public Gridworld(String id, String opponent, int steps, int gsizex, int gsizey,
			int corralx0, int corraly0, int corralx1, int corraly1, int lineOfSight) {
		this.id = id;
		this.opponent = opponent;
		this.steps = steps;
		this.gsizex = gsizex;
		this.gsizey = gsizey;
		this.corralx0 = corralx0;
		this.corraly0 = corraly0;
		this.corralx1 = corralx1;
		this.corraly1 = corraly1;
		this.lineOfSight = lineOfSight;
		
		grid = new GridField[gsizex][gsizey];
		for (int x = 0; x < gsizex; x++) {
			for (int y = 0; y < gsizey; y++) {
				if (isCorralField(x,y)){
					grid[x][y] = new GridField(x,y,IMAPCConstants.TYPE_CORRAL,-1);
				}else{
					grid[x][y] = new GridField(x,y,IMAPCConstants.TYPE_UNKNOWN,-1);
				}
			}
		}
		switches = new IntHashMap<Switch>();
	}
	
	public Gridworld() {
		// empty constructor is needed for desirialization!
		switches = new IntHashMap<Switch>();
	}
	
	/**
	 * checks, whether a field is a fence field
	 * @param field the field
	 * @return true, if it is a fence or openfence
	 */
	private boolean isFence(GridField field){
		return (field.getType() == IMAPCConstants.TYPE_FENCE || field.getType() == IMAPCConstants.TYPE_OPEN_FENCE);
	}
	
	/**
	 * searches the switch for a given fenceField
	 * @param fenceField the fenceField
	 * @return the switch
	 *  can be null, if no switch was found or fenceField was no fence
	 */
	public Switch findSwitchForFence(Position fenceField) {
		
		if(!isFence(grid[fenceField.getX()][fenceField.getY()])){
			return null;
		}
		int x = fenceField.getX();
		int y = fenceField.getY();
		// search west
		while(isFence(grid[x][y])){
			x--;
		}
		if(grid[x][y].getType() == IMAPCConstants.TYPE_SWITCH){
			return switches.get(Position.getIDForXY(x, y));
		}
		x = fenceField.getX();
		y = fenceField.getY();
		// search east
		while(isFence(grid[x][y])){
			x++;
		}
		if(grid[x][y].getType() == IMAPCConstants.TYPE_SWITCH){
			return switches.get(Position.getIDForXY(x, y));
		}
		x = fenceField.getX();
		y = fenceField.getY();
		// search north
		while(isFence(grid[x][y])){
			y--;
		}
		if(grid[x][y].getType() == IMAPCConstants.TYPE_SWITCH){
			return switches.get(Position.getIDForXY(x, y));
		}
		x = fenceField.getX();
		y = fenceField.getY();
		// search south
		while(isFence(grid[x][y])){
			y++;
		}
		if(grid[x][y].getType() == IMAPCConstants.TYPE_SWITCH){
			return switches.get(Position.getIDForXY(x, y));
		}
		return null;
	}
	
	
	
	public GridField[][] getGrid(){
		return this.grid;
	}
	
	public int getSteps(){
		return this.steps;
	}
	
	public int getGsizeX(){
		return this.gsizex;
	}
	
	public int getGsizeY(){
		return this.gsizey;
	}
	
	public int getCorralX0(){
		return this.corralx0;
	}
	
	public int getCorralX1(){
		return this.corralx1;
	}
	
	public int getCorralY0(){
		return this.corraly0;
	}
	
	public int getCorralY1(){
		return this.corraly1;
	}
	
	public int getLineOfSight(){
		return this.lineOfSight;
	}

	public IntHashMap<Switch> getSwitches() {
		return switches;
	}

	public void setSwitches(IntHashMap<Switch> switches) {
		this.switches = switches;
	}

	/**
	 * checks, wheter a field with given coordinates is part of the own corral
	 * @param x the x coordinate of the field
	 * @param y the y coordinate of the field
	 * @return true, if the field is part of the corral
	 */
	private boolean isCorralField(int x, int y){
		if (x >= corralx0 && x <= corralx1 && y >= corraly0 && y <= corraly1){
			// field is a corralField
			return true;
		}
		return false;
	}

	/**
	 * updates the gridworld with an percetion
	 * @param gridInfo the fields seen by an Agent
	 */
	public void update(GridField[] gridInfo){
		for(GridField f : gridInfo){
			if (f == null || f.getType() == IMAPCConstants.TYPE_UNKNOWN) continue; //dont write unknown fields into the grid
			if (f.getType() == IMAPCConstants.TYPE_SWITCH){
				// found switch, check, wheter it is already in the list
				if(switches.get(f.getId()) == null){
					// not in the list, insert
					Switch s = new Switch(f.getX(), f.getY());
					switches.put(s.getId(), s);
				}
			}
			grid[f.getX()][f.getY()] = f;
		}
	}
	
	/**
	 * luftweg zwischen zwei punkten, zur entfernungsschaetzung
	 * @param start
	 * @param destination
	 * @return
	 */
	public int distance(Position start, Position destination){
		return Math.max(Math.abs(start.getX()-destination.getX()), Math.abs(start.getY()-destination.getY()));
	}	
	
	@Override
	public void deserialise(IDeserialiser arg0) throws IOException,
			ClassNotFoundException {
		this.id = arg0.readString();
		this.opponent = arg0.readString();
		this.steps = arg0.readInt();
		this.gsizex = arg0.readInt();
		this.gsizey = arg0.readInt();
		this.corralx0 = arg0.readInt();
		this.corraly0 = arg0.readInt();
		this.corralx1 = arg0.readInt();
		this.corraly1 = arg0.readInt();
		this.lineOfSight = arg0.readInt();
		int sizex = arg0.readInt();
		int sizey = arg0.readInt();
		this.grid = new GridField[sizex][sizey];
		for(int x = 0; x < grid.length; x++){
			for(int y = 0; y < grid[0].length; y++){
				grid[x][y] = (GridField)arg0.readObject();
			}
		}
		int size = arg0.readInt();
		for (int i = 0; i < size; i++) {
			Switch s = (Switch) arg0.readObject();
			switches.put(s.getId(), s);
		}
	}

	@Override
	public void serialise(ISerialiser arg0) throws IOException {
		arg0.writeString(id);
		arg0.writeString(opponent);
		arg0.writeInt(steps);
		arg0.writeInt(gsizex);
		arg0.writeInt(gsizey);
		arg0.writeInt(corralx0);
		arg0.writeInt(corraly0);
		arg0.writeInt(corralx1);
		arg0.writeInt(corraly1);
		arg0.writeInt(lineOfSight);
		arg0.writeInt(grid.length);
		arg0.writeInt(grid[0].length);
		for(int x = 0; x < grid.length; x++){
			for(int y = 0; y < grid[0].length; y++){
				arg0.writeObject(grid[x][y]);
			}
		}
		arg0.writeInt(switches.size());
		Object[] values =  switches.valuesAsArray();
		for (Object value : values){
			arg0.writeObject(value);
		}
	}

}
