package de.tub.mapc.parser;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.kxml2.io.KXmlParser;
import org.kxml2.kdom.Document;
import org.kxml2.kdom.Element;
import org.kxml2.kdom.Node;
import org.slf4j.Logger;
import org.xmlpull.v1.XmlPullParserException;

import de.jiac.micro.agent.memory.IFact;
import de.jiac.micro.ext.io.ByteArrayInputBuffer;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.knowledge.AuthenticationResult;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.knowledge.Perception;
import de.tub.mapc.knowledge.SimulationEnd;
import de.tub.mapc.worldmodel.GridField;

/**
 * This class parses the XML messages and converts into the agent knowledge.
 * 
 */
@SuppressWarnings("unused")
public class MAPCMessageParser {

	/** The agent name. */	
	private String _agentName;

	/** The agent id. */
	private int _agentID;

	/** The parser. */
	private KXmlParser _parser;

	/** The logger */
	private Logger _logger;

	/**
	 * Instantiates a new mAPC message parser.
	 * 
	 * @param agentName
	 *            the agent name
	 * @param agentID
	 *            the agent id
	 */
	public MAPCMessageParser(Logger logger, String agentName, int agentID) {
		_logger = logger;
		_agentName = agentName;
		_agentID = agentID;
		_parser = new KXmlParser();
	}

	/**
	 * Parses the received XML.
	 * 
	 * @param raw
	 *            the raw
	 * 
	 * @return the i fact
	 * 
	 * @throws XmlPullParserException
	 *             the XML pull parser exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	public IFact parseXML(byte[] raw) throws XmlPullParserException, IOException {
		if (raw == null || raw.length == 0)
			return null;
		Document document = createDocument(raw);
		Element message = document.getRootElement();
		if (message == null)
			return null;

		long timestamp = -1;
		try {
			timestamp = Long.parseLong(message.getAttributeValue("", "timestamp"));
		} catch (NumberFormatException e) {
			// ignore it
		}
		// System.out.println(timestamp);

		if (message.getName().equalsIgnoreCase("message")) {
			String type = message.getAttributeValue("", "type");
			if (type.trim().equalsIgnoreCase("request-action")) {
				return createPerception(message);
			} else if (type.trim().equalsIgnoreCase("auth-response")) {
				return createAuthenticationResponse(message);
			} else if (type.trim().equalsIgnoreCase("sim-start")) {
				return createSimulationStart(message);
			} else if (type.trim().equalsIgnoreCase("sim-end")) {
				return createSimulationEnd(message);
			} else if (type.trim().equalsIgnoreCase("bye")) {
				return createBye(message);
			} 
		}
		return null;
	}

	/**
	 * Creates the document.
	 * 
	 * @param raw
	 *            the raw
	 * 
	 * @return the document
	 * 
	 * @throws XmlPullParserException
	 *             the xml pull parser exception
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	private Document createDocument(byte[] raw) throws XmlPullParserException, IOException {
		Document document = new Document();
		ByteArrayInputBuffer buffer = new ByteArrayInputBuffer();
		buffer.setInput(raw);
		// _parser.setInput(_xmlInputBuffer, "UTF-8");
		_parser.setInput(buffer, "ISO-8859-1");
		document.parse(_parser);
		return document;
	}

	/**
	 * Creates the authentication response.
	 * 
	 * @param message
	 *            the message
	 * 
	 * @return the authentication response
	 */
	private IFact createAuthenticationResponse(Element message) {
		Element authresult = getElementChild(message, "authentication");
		if (authresult != null) {
			String result = authresult.getAttributeValue("", "result");
			//_logger.info("authentication result received: " + result);
			return new AuthenticationResult(result);
		}		
		return null;
	}

	/**
	 * Creates the simulation start.
	 * 
	 * @param message
	 *            the message
	 * 
	 * @return the simulation start
	 */
	private IFact createSimulationStart(Element message) {
		Element simstart = getElementChild(message, "simulation");
		if (simstart != null) {
			String id = simstart.getAttributeValue("", "id");
			String opponent = simstart.getAttributeValue("", "opponent");
			int steps = Integer.parseInt(simstart.getAttributeValue("", "steps"));
			int gsizex = Integer.parseInt(simstart.getAttributeValue("", "gsizex"));
			int gsizey = Integer.parseInt(simstart.getAttributeValue("", "gsizey"));
			int corralx0 = Integer.parseInt(simstart.getAttributeValue("", "corralx0"));
			int corraly0 = Integer.parseInt(simstart.getAttributeValue("", "corraly0"));
			int corralx1 = Integer.parseInt(simstart.getAttributeValue("", "corralx1"));
			int corraly1 = Integer.parseInt(simstart.getAttributeValue("", "corraly1"));
			int lineOfSight = Integer.parseInt(simstart.getAttributeValue("", "lineOfSight"));
			//_logger.info("Simulation start message received with data:" + " id=" + id + " opponent=" + opponent + " steps=" + steps + " gridSizeX=" + gsizex + " gridSizeY=" + gsizey + " corralX0=" + corralx0 + " corralY0=" + corraly0 + " corralX1=" + corralx1 + " corralY1=" + corraly1 + " lineOfSight=" + lineOfSight);
						
			return new Gridworld(id, opponent, steps, gsizex, gsizey, corralx0, corraly0, corralx1, corraly1, lineOfSight);
		}
		return null;
	}

	/**
	 * Creates the simulation end.
	 * 
	 * @param message
	 *            the message
	 * 
	 * @return the simulation end
	 */
	private IFact createSimulationEnd(Element message) {
		Element simend = getElementChild(message, "sim-result");
		if (simend != null) {
			String result = simend.getAttributeValue("", "result");
			float score = Float.parseFloat(simend.getAttributeValue("", "averageScore"));
			//_logger.info("Simulation end message received: result=" + result + " score=" + score);
			return new SimulationEnd(result, score);
		}
		return null;
	}

	/**
	 * Creates the perception.
	 * 
	 * @param message
	 *            the message
	 * 
	 * @return the perception
	 */
	private IFact createPerception(Element message) {
		Element perception = getElementChild(message, "perception");
		if (perception != null) {
			int step = Integer.parseInt(perception.getAttributeValue("", "step"));
			short posx = Short.parseShort(perception.getAttributeValue("", "posx"));
			short posy = Short.parseShort(perception.getAttributeValue("", "posy"));
			int score = Integer.parseInt(perception.getAttributeValue("", "cowsInCorral"));
			long deadline = Long.parseLong(perception.getAttributeValue("", "deadline"));
			String id = perception.getAttributeValue("", "id");
			Element[] array = getCellElements(perception);

			StringBuffer cellInfo = new StringBuffer();
			GridField[] gridInfo = new GridField[array.length];
			for (int i = 0; i < array.length; i++) {
				Element cell = array[i];
				// the coordinates are relative to the position of the Agent
				int x = Integer.parseInt(cell.getAttributeValue("", "x"));
				int y = Integer.parseInt(cell.getAttributeValue("", "y"));
				// get the type of the cell
				int type = getCellInfo(cell);
				
				gridInfo[i] = new GridField(posx+x, posy+y, type, step);
			}

			//_logger.info("Perception received:" + " step=" + step + " posx=" + posx + " posy=" + posy + " score=" + score + " deadline=" + deadline + " cells=\n" + cellInfo.toString() + "\n");
			return new Perception(step, posx, posy, score, deadline, id, gridInfo);
		}
		return null;
	}

	/**
	 * Gets the cell info.
	 * 
	 * @param cell
	 *            the cell
	 * @param x
	 *            the x
	 * @param y
	 *            the y
	 */
	private void printCellInfo(Element cell, int x, int y, StringBuffer cellInfo) {
		cellInfo.append("\tcell(" + x + ", " + y + ") is/has ");
		for (int j = 0; j < cell.getChildCount(); j++) {
			if (cell.getType(j) == Node.ELEMENT) {
				Element temp = (Element) cell.getChild(j);				
				
				if (temp.getName().equalsIgnoreCase("empty")) {
					cellInfo.append("empty.");
				} else if (temp.getName().equalsIgnoreCase("obstacle")) {
					cellInfo.append("an obstacle.");
				} else if (temp.getName().equalsIgnoreCase("cow")) {
					int cowId = Integer.parseInt(temp.getAttributeValue("", "ID"));
					cellInfo.append("cow with ID=" + cowId + ".");
				} else if (temp.getName().equalsIgnoreCase("corral")) {
					if (temp.getAttributeValue("", "type").equalsIgnoreCase("enemy")) {
						cellInfo.append("enemy corral.");
					} else {
						cellInfo.append("own corral.");
					}
				} else if (temp.getName().equalsIgnoreCase("agent")) {
					if (temp.getAttributeValue("", "type").equals("ally")) {
						cellInfo.append("an ally agent.");
					} else {
						cellInfo.append("an enemy agent.");
					}
				} else if (temp.getName().equalsIgnoreCase("switch")) {
					cellInfo.append("switch.");
				} else if (temp.getName().equalsIgnoreCase("unknown")) {
					cellInfo.append("unknown.");
				} else if (temp.getName().equalsIgnoreCase("fence")) {
					if (temp.getAttributeValue("", "open").equalsIgnoreCase("false")) {
						cellInfo.append("closed fence.");
					} else {
						cellInfo.append("open corral.");
					}
				} else {
					System.err.println(("error... unknown perception..."));
					int attributcount = temp.getAttributeCount();
					System.err.print("attributcount: " + attributcount + ", ");
					System.err.println("tempName: " + temp.getName());
					System.err.println("tempName: " + temp.getName() + ", attributcount: " + attributcount + " ");
					for (int i = 0; i < attributcount; i++) {
						System.err.print("AttributValue: " + temp.getAttributeValue("", "open") + ", ");
						System.err.print(" AttributName " + i + ": " + temp.getAttributeName(i) + " ");
					}
				}
			}
		}
		cellInfo.append('\n');
	}

	/**
	 * Gets the cell info.
	 * 
	 * @param cell
	 *            the cell
	 * @param x
	 *            the x
	 * @param y
	 *            the y
	 */
	private int getCellInfo(Element cell) {
		for (int j = 0; j < cell.getChildCount(); j++) {
			if (cell.getType(j) == Node.ELEMENT) {
				Element temp = (Element) cell.getChild(j);				
				
				if (temp.getName().equalsIgnoreCase("empty")) {
					return IMAPCConstants.TYPE_EMPTY;
				} else if (temp.getName().equalsIgnoreCase("obstacle")) {
					return IMAPCConstants.TYPE_TREE;
				} else if (temp.getName().equalsIgnoreCase("cow")) {
					int cowId = Integer.parseInt(temp.getAttributeValue("", "ID"));
					//TODO was ist mit der cowid? brauchen wir die?
					return IMAPCConstants.TYPE_COW;
				} else if (temp.getName().equalsIgnoreCase("corral")) {
					if (temp.getAttributeValue("", "type").equalsIgnoreCase("enemy")) {
						return IMAPCConstants.TYPE_ENEMY_CORRAL;
					} else {
						return IMAPCConstants.TYPE_CORRAL;
					}
				} else if (temp.getName().equalsIgnoreCase("agent")) {
					if (temp.getAttributeValue("", "type").equals("ally")) {
						return IMAPCConstants.TYPE_PARTNER;
					} else {
						return IMAPCConstants.TYPE_ENEMY;
					}
				} else if (temp.getName().equalsIgnoreCase("switch")) {
					return IMAPCConstants.TYPE_SWITCH;
				} else if (temp.getName().equalsIgnoreCase("unknown")) {
					return IMAPCConstants.TYPE_UNKNOWN;
				} else if (temp.getName().equalsIgnoreCase("fence")) {
					if (temp.getAttributeValue("", "open").equalsIgnoreCase("false")) {
						return IMAPCConstants.TYPE_FENCE;
					} else {
						return IMAPCConstants.TYPE_OPEN_FENCE;
					}
				} else {
					System.err.println(("error... unknown perception..."));
					int attributcount = temp.getAttributeCount();
					System.err.print("attributcount: " + attributcount + ", ");
					System.err.println("tempName: " + temp.getName());
					System.err.println("tempName: " + temp.getName() + ", attributcount: " + attributcount + " ");
					for (int i = 0; i < attributcount; i++) {
						System.err.print("AttributValue: " + temp.getAttributeValue("", "open") + ", ");
						System.err.print(" AttributName " + i + ": " + temp.getAttributeName(i) + " ");
					}
				}
			}
		}
		return IMAPCConstants.TYPE_UNKNOWN;
	}
	
	
	/**
	 * Creates the bye.
	 * 
	 * @param message
	 *            the message
	 * 
	 * @return the bye
	 */
	private IFact createBye(Element message) {
		//_logger.info("Bye message received");
		// TODO: create and return the model for the "bye" message.
		return null;
	}

	/**
	 * Gets the cell elements.
	 * 
	 * @param parent
	 *            the parent
	 * 
	 * @return the cell elements
	 */
	private Element[] getCellElements(Element parent) {
		List<Element> list = new ArrayList<Element>();
		int cellCount = parent.getChildCount();

		for (int i = 0; i < cellCount; i++) {
			if (parent.getType(i) == Node.ELEMENT) {
				list.add((Element) parent.getChild(i));

			}
		}
		return list.toArray(new Element[0]);
	}

	/**
	 * Gets the element child.
	 * 
	 * @param parent
	 *            the parent
	 * @param elementName
	 *            the element name
	 * 
	 * @return the element child
	 */
	private Element getElementChild(Element parent, String elementName) {
		Element child = null;
		for (int i = 0; i < parent.getChildCount(); i++) {
			if (parent.getType(i) == Node.ELEMENT) {
				Element temp = (Element) parent.getChild(i);
				if (temp.getName().equalsIgnoreCase(elementName)) {
					child = temp;
					break;
				}
			}
		}
		return child;
	}

}
